// This file is under GNU General Public License 3.0
// see LICENSE.txt

#include "KeySync_fsm.h"
#include <stdlib.h>

#ifdef NDEBUG
static
#endif
const char *KeySync_state_name(int state)
{
    switch (state) {
        case End:
            return "End";
        case None:
            return "None";
        case Init:
            return "InitState";
        case Sole:
            return "Sole";
        case HandshakingOfferer:
            return "HandshakingOfferer";
        case HandshakingRequester:
            return "HandshakingRequester";
        case HandshakingPhase1Offerer:
            return "HandshakingPhase1Offerer";
        case HandshakingPhase1Requester:
            return "HandshakingPhase1Requester";
        case HandshakingPhase2Offerer:
            return "HandshakingPhase2Offerer";
        case HandshakingPhase2Requester:
            return "HandshakingPhase2Requester";
        case FormingGroupOfferer:
            return "FormingGroupOfferer";
        case FormingGroupRequester:
            return "FormingGroupRequester";
        case Grouped:
            return "Grouped";
        case HandshakingToJoin:
            return "HandshakingToJoin";
        case HandshakingToJoinPhase1:
            return "HandshakingToJoinPhase1";
        case HandshakingToJoinPhase2:
            return "HandshakingToJoinPhase2";
        case JoiningGroup:
            return "JoiningGroup";
        case HandshakingGrouped:
            return "HandshakingGrouped";
        case HandshakingGroupedPhase1:
            return "HandshakingGroupedPhase1";
        case HandshakingGroupedPhase2:
            return "HandshakingGroupedPhase2";
        default:
            return "unknown state";
    }
}

#ifdef NDEBUG
static
#endif
const char *KeySync_event_name(int event)
{
    switch (event) {
        case None:
            return "None";
        case Init:
            return "Init";
        case KeyGen:
            return "KeyGen";
        case CannotDecrypt:
            return "CannotDecrypt";
        case Beacon:
            return "Beacon";
        case NegotiationRequest:
            return "NegotiationRequest";
        case NegotiationOpen:
            return "NegotiationOpen";
        case Cancel:
            return "Cancel";
        case Rollback:
            return "Rollback";
        case Reject:
            return "Reject";
        case CommitReject:
            return "CommitReject";
        case Accept:
            return "Accept";
        case CommitAcceptRequester:
            return "CommitAcceptRequester";
        case CommitAcceptOfferer:
            return "CommitAcceptOfferer";
        case OwnKeysRequester:
            return "OwnKeysRequester";
        case OwnKeysOfferer:
            return "OwnKeysOfferer";
        case GroupKeys:
            return "GroupKeys";
        case GroupTrustThisKey:
            return "GroupTrustThisKey";
        case CommitAcceptForGroup:
            return "CommitAcceptForGroup";
        case CommitAccept:
            return "CommitAccept";
        default:
            return "unknown event";
    }
}


static char *_str(int n, bool hex)
{
    char *buf = calloc(1, 24);
    assert(buf);
    if (!buf)
        return NULL;

    if (hex)
        snprintf(buf, 24, "%.4x", n);
    else
        snprintf(buf, 24, "%d", n);
    return buf;
}

#define KeySync_ERR_LOG(t, d) log_event(session, (t), "KeySync", (d), "error")

static PEP_STATUS _KeySync_ERR_LOG_int(PEP_SESSION session, char *t, int n, bool hex)
{
    char *_buf = _str(n, hex);
    if (!_buf)
        return PEP_OUT_OF_MEMORY;
    PEP_STATUS status = KeySync_ERR_LOG(t, _buf);
    free(_buf);
    return status;
}

#define KeySync_ERR_LOG_INT(t, n) _KeySync_ERR_LOG_int(session, (t), (n), false)
#define KeySync_ERR_LOG_HEX(t, n) _KeySync_ERR_LOG_int(session, (t), (n), true)
#define KeySync_SERVICE_LOG(t, d) SERVICE_LOG(session, (t), "KeySync", (d))

KeySync_state fsm_KeySync(
        PEP_SESSION session,
        KeySync_state state,
        KeySync_event event
    )
{
    assert(session);
    if (!session)
        return invalid_state;

    if (state == None)
        state = KeySync_state_Init;

    switch (state) {
        case KeySync_state_Init:
            KeySync_SERVICE_LOG("in state", "InitState");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    status = deviceGrouped(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition deviceGrouped failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "deviceGrouped");
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case Sole:
            KeySync_SERVICE_LOG("in state", "Sole");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "newChallengeAndNegotiationBase");
                    status = newChallengeAndNegotiationBase(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action newChallengeAndNegotiationBase() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showBeingSole");
                    status = showBeingSole(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showBeingSole() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "Beacon");
                    status = send_Sync_message(session, 1, KeySync_PR_beacon);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Beacon failed", status);
                        return cannot_send;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                case KeyGen: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "KeyGen");
                    
                    KeySync_SERVICE_LOG("send message", "Beacon");
                    status = send_Sync_message(session, 1, KeySync_PR_beacon);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Beacon failed", status);
                        return cannot_send;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                case CannotDecrypt: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "CannotDecrypt");
                    
                    KeySync_SERVICE_LOG("send message", "Beacon");
                    status = send_Sync_message(session, 1, KeySync_PR_beacon);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Beacon failed", status);
                        return cannot_send;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                case Beacon: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Beacon");
                    
                    status = sameChallenge(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameChallenge failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameChallenge");
                    }
                    else {
                        KeySync_SERVICE_LOG("condition does not apply", "sameChallenge");
                        
                        status = weAreOfferer(session, &result);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("computing condition weAreOfferer failed", status);
                            return invalid_condition;
                        }
                        if (result) {
                            KeySync_SERVICE_LOG("condition applies", "weAreOfferer");
                            
                            KeySync_SERVICE_LOG("do action", "useOwnChallenge");
                            status = useOwnChallenge(session);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("executing action useOwnChallenge() failed", status);
                                return invalid_action;
                            }
                            
                            KeySync_SERVICE_LOG("send message", "Beacon");
                            status = send_Sync_message(session, 1, KeySync_PR_beacon);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("sending Beacon failed", status);
                                return cannot_send;
                            }
                        }
                        else {
                            KeySync_SERVICE_LOG("condition does not apply", "weAreOfferer");
                            
                            KeySync_SERVICE_LOG("do action", "openNegotiation");
                            status = openNegotiation(session);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("executing action openNegotiation() failed", status);
                                return invalid_action;
                            }
                            
                            KeySync_SERVICE_LOG("do action", "tellWeAreNotGrouped");
                            status = tellWeAreNotGrouped(session);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("executing action tellWeAreNotGrouped() failed", status);
                                return invalid_action;
                            }
                            
                            KeySync_SERVICE_LOG("send message", "NegotiationRequest");
                            status = send_Sync_message(session, 1, KeySync_PR_negotiationRequest);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("sending NegotiationRequest failed", status);
                                return cannot_send;
                            }
                            
                            KeySync_SERVICE_LOG("do action", "useOwnChallenge");
                            status = useOwnChallenge(session);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("executing action useOwnChallenge() failed", status);
                                return invalid_action;
                            }
                        }
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                case NegotiationRequest: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "NegotiationRequest");
                    
                    status = sameChallenge(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameChallenge failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameChallenge");
                        
                        status = sameNegotiation(session, &result);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("computing condition sameNegotiation failed", status);
                            return invalid_condition;
                        }
                        if (result) {
                            KeySync_SERVICE_LOG("condition applies", "sameNegotiation");
                        }
                        else {
                            KeySync_SERVICE_LOG("condition does not apply", "sameNegotiation");
                            
                            KeySync_SERVICE_LOG("do action", "storeNegotiation");
                            status = storeNegotiation(session);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("executing action storeNegotiation() failed", status);
                                return invalid_action;
                            }
                            
                            KeySync_SERVICE_LOG("send message", "NegotiationOpen");
                            status = send_Sync_message(session, 1, KeySync_PR_negotiationOpen);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("sending NegotiationOpen failed", status);
                                return cannot_send;
                            }
                            
                            status = partnerIsGrouped(session, &result);
                            if (status == PEP_OUT_OF_MEMORY)
                                return out_of_memory;
                            if (status) {
                                KeySync_ERR_LOG_HEX("computing condition partnerIsGrouped failed", status);
                                return invalid_condition;
                            }
                            if (result) {
                                KeySync_SERVICE_LOG("condition applies", "partnerIsGrouped");
                                
                                KeySync_SERVICE_LOG("transition to state", "HandshakingToJoin");
                                return HandshakingToJoin;
                            }
                            else {
                                KeySync_SERVICE_LOG("condition does not apply", "storeNegotiation");
                                
                                KeySync_SERVICE_LOG("transition to state", "HandshakingOfferer");
                                return HandshakingOfferer;
                            }
                        }
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                case NegotiationOpen: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "NegotiationOpen");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "storeNegotiation");
                        status = storeNegotiation(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action storeNegotiation() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingRequester");
                        return HandshakingRequester;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Sole");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingOfferer:
            KeySync_SERVICE_LOG("in state", "HandshakingOfferer");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "showSoleHandshake");
                    status = showSoleHandshake(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showSoleHandshake() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingOfferer");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingOfferer");
                    break;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingOfferer");
                    break;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "CommitAcceptOfferer");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAcceptOfferer);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAcceptOfferer failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "HandshakingPhase1Offerer");
                    return HandshakingPhase1Offerer;
                }
                
                case CommitAcceptRequester: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptRequester");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingPhase2Offerer");
                        return HandshakingPhase2Offerer;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingOfferer");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingRequester:
            KeySync_SERVICE_LOG("in state", "HandshakingRequester");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "showSoleHandshake");
                    status = showSoleHandshake(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showSoleHandshake() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingRequester");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingRequester");
                    break;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingRequester");
                    break;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "CommitAcceptRequester");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAcceptRequester);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAcceptRequester failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "HandshakingPhase1Requester");
                    return HandshakingPhase1Requester;
                }
                
                case CommitAcceptOfferer: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptOfferer");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingPhase2Requester");
                        return HandshakingPhase2Requester;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingRequester");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingPhase1Offerer:
            KeySync_SERVICE_LOG("in state", "HandshakingPhase1Offerer");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "untrustThisKey");
                        status = untrustThisKey(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action untrustThisKey() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Offerer");
                    break;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "untrustThisKey");
                        status = untrustThisKey(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action untrustThisKey() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Offerer");
                    break;
                }
                
                case CommitAcceptRequester: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptRequester");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "FormingGroupOfferer");
                        return FormingGroupOfferer;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Offerer");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingPhase1Requester:
            KeySync_SERVICE_LOG("in state", "HandshakingPhase1Requester");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "untrustThisKey");
                        status = untrustThisKey(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action untrustThisKey() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Requester");
                    break;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "untrustThisKey");
                        status = untrustThisKey(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action untrustThisKey() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Requester");
                    break;
                }
                
                case CommitAcceptOfferer: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptOfferer");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "FormingGroupRequester");
                        return FormingGroupRequester;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingPhase1Requester");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingPhase2Offerer:
            KeySync_SERVICE_LOG("in state", "HandshakingPhase2Offerer");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("send message", "CommitAcceptOfferer");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAcceptOfferer);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAcceptOfferer failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "FormingGroupOfferer");
                    return FormingGroupOfferer;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingPhase2Requester:
            KeySync_SERVICE_LOG("in state", "HandshakingPhase2Requester");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("send message", "CommitAcceptRequester");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAcceptRequester);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAcceptRequester failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "FormingGroupRequester");
                    return FormingGroupRequester;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case FormingGroupOfferer:
            KeySync_SERVICE_LOG("in state", "FormingGroupOfferer");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                    status = prepareOwnKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "OwnKeysOfferer");
                    status = send_Sync_message(session, 1, KeySync_PR_ownKeysOfferer);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending OwnKeysOfferer failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showFormingGroup");
                    status = showFormingGroup(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showFormingGroup() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "FormingGroupOfferer");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Rollback: {
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case OwnKeysRequester: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "OwnKeysRequester");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "receivedKeysAreDefaultKeys");
                    status = receivedKeysAreDefaultKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action receivedKeysAreDefaultKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showGroupCreated");
                    status = showGroupCreated(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showGroupCreated() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case FormingGroupRequester:
            KeySync_SERVICE_LOG("in state", "FormingGroupRequester");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "showFormingGroup");
                    status = showFormingGroup(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showFormingGroup() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "FormingGroupRequester");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Rollback: {
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case OwnKeysOfferer: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "OwnKeysOfferer");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                    status = prepareOwnKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "ownKeysAreDefaultKeys");
                    status = ownKeysAreDefaultKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action ownKeysAreDefaultKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "OwnKeysRequester");
                    status = send_Sync_message(session, 1, KeySync_PR_ownKeysRequester);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending OwnKeysRequester failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showGroupCreated");
                    status = showGroupCreated(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showGroupCreated() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case Grouped:
            KeySync_SERVICE_LOG("in state", "Grouped");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "newChallengeAndNegotiationBase");
                    status = newChallengeAndNegotiationBase(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action newChallengeAndNegotiationBase() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showBeingInGroup");
                    status = showBeingInGroup(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showBeingInGroup() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                case GroupKeys: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupKeys");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                case KeyGen: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "KeyGen");
                    
                    KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                    status = prepareOwnKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "GroupKeys");
                    status = send_Sync_message(session, 1, KeySync_PR_groupKeys);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending GroupKeys failed", status);
                        return cannot_send;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                case Beacon: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Beacon");
                    
                    KeySync_SERVICE_LOG("do action", "openNegotiation");
                    status = openNegotiation(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action openNegotiation() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "tellWeAreGrouped");
                    status = tellWeAreGrouped(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action tellWeAreGrouped() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "NegotiationRequest");
                    status = send_Sync_message(session, 1, KeySync_PR_negotiationRequest);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending NegotiationRequest failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "useOwnChallenge");
                    status = useOwnChallenge(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action useOwnChallenge() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                case NegotiationOpen: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "NegotiationOpen");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "storeNegotiation");
                        status = storeNegotiation(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action storeNegotiation() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingGrouped");
                        return HandshakingGrouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                case GroupTrustThisKey: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupTrustThisKey");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "Grouped");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingToJoin:
            KeySync_SERVICE_LOG("in state", "HandshakingToJoin");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "showJoinGroupHandshake");
                    status = showJoinGroupHandshake(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showJoinGroupHandshake() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoin");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoin");
                    break;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case CommitAcceptForGroup: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptForGroup");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingToJoinPhase2");
                        return HandshakingToJoinPhase2;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoin");
                    break;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoin");
                    break;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "CommitAccept");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAccept);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAccept failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "HandshakingToJoinPhase1");
                    return HandshakingToJoinPhase1;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingToJoinPhase1:
            KeySync_SERVICE_LOG("in state", "HandshakingToJoinPhase1");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Sole");
                        return Sole;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoinPhase1");
                    break;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "disable");
                        status = disable(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "End");
                        return End;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoinPhase1");
                    break;
                }
                
                case CommitAcceptForGroup: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAcceptForGroup");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "JoiningGroup");
                        return JoiningGroup;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingToJoinPhase1");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingToJoinPhase2:
            KeySync_SERVICE_LOG("in state", "HandshakingToJoinPhase2");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Sole");
                    return Sole;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "disable");
                    status = disable(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action disable() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "End");
                    return End;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "JoiningGroup");
                    return JoiningGroup;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case JoiningGroup:
            KeySync_SERVICE_LOG("in state", "JoiningGroup");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case GroupKeys: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupKeys");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "receivedKeysAreDefaultKeys");
                    status = receivedKeysAreDefaultKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action receivedKeysAreDefaultKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                    status = prepareOwnKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "GroupKeys");
                    status = send_Sync_message(session, 1, KeySync_PR_groupKeys);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending GroupKeys failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "showDeviceAdded");
                    status = showDeviceAdded(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showDeviceAdded() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingGrouped:
            KeySync_SERVICE_LOG("in state", "HandshakingGrouped");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Init");
                    
                    KeySync_SERVICE_LOG("do action", "showGroupedHandshake");
                    status = showGroupedHandshake(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action showGroupedHandshake() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "GroupTrustThisKey");
                    status = send_Sync_message(session, 1, KeySync_PR_groupTrustThisKey);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending GroupTrustThisKey failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "CommitAcceptForGroup");
                    status = send_Sync_message(session, 1, KeySync_PR_commitAcceptForGroup);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitAcceptForGroup failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "HandshakingGroupedPhase1");
                    return HandshakingGroupedPhase1;
                }
                
                case CommitAccept: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAccept");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "HandshakingGroupedPhase2");
                        return HandshakingGroupedPhase2;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                case GroupTrustThisKey: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupTrustThisKey");
                    
                    KeySync_SERVICE_LOG("do action", "hideHandshakeDialog");
                    status = hideHandshakeDialog(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action hideHandshakeDialog() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                case GroupKeys: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupKeys");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGrouped");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingGroupedPhase1:
            KeySync_SERVICE_LOG("in state", "HandshakingGroupedPhase1");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Rollback: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "Rollback");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase1");
                    break;
                }
                
                case CommitReject: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitReject");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase1");
                    break;
                }
                
                case CommitAccept: {
                    PEP_STATUS status;
                    bool result = false;
                
                    KeySync_SERVICE_LOG("received event", "CommitAccept");
                    
                    status = sameNegotiationAndPartner(session, &result);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("computing condition sameNegotiationAndPartner failed", status);
                        return invalid_condition;
                    }
                    if (result) {
                        KeySync_SERVICE_LOG("condition applies", "sameNegotiationAndPartner");
                        
                        KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                        status = prepareOwnKeys(session);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                            return invalid_action;
                        }
                        
                        KeySync_SERVICE_LOG("send message", "GroupKeys");
                        status = send_Sync_message(session, 1, KeySync_PR_groupKeys);
                        if (status == PEP_OUT_OF_MEMORY)
                            return out_of_memory;
                        if (status) {
                            KeySync_ERR_LOG_HEX("sending GroupKeys failed", status);
                            return cannot_send;
                        }
                        
                        KeySync_SERVICE_LOG("transition to state", "Grouped");
                        return Grouped;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase1");
                    break;
                }
                
                case GroupTrustThisKey: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupTrustThisKey");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase1");
                    break;
                }
                
                case GroupKeys: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupKeys");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase1");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        case HandshakingGroupedPhase2:
            KeySync_SERVICE_LOG("in state", "HandshakingGroupedPhase2");
        
            switch (event) {
                case None:
                    KeySync_SERVICE_LOG("received None event", "ignoring");
                    break;
        
                case Init:
                    KeySync_SERVICE_LOG("received Init but nothing to do", "Init");
                    break;
        
                case Cancel: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Cancel");
                    
                    KeySync_SERVICE_LOG("send message", "Rollback");
                    status = send_Sync_message(session, 1, KeySync_PR_rollback);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending Rollback failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                case Reject: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Reject");
                    
                    KeySync_SERVICE_LOG("send message", "CommitReject");
                    status = send_Sync_message(session, 1, KeySync_PR_commitReject);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending CommitReject failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                case Accept: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "Accept");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "GroupTrustThisKey");
                    status = send_Sync_message(session, 1, KeySync_PR_groupTrustThisKey);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending GroupTrustThisKey failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("do action", "prepareOwnKeys");
                    status = prepareOwnKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action prepareOwnKeys() failed", status);
                        return invalid_action;
                    }
                    
                    KeySync_SERVICE_LOG("send message", "GroupKeys");
                    status = send_Sync_message(session, 1, KeySync_PR_groupKeys);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("sending GroupKeys failed", status);
                        return cannot_send;
                    }
                    
                    KeySync_SERVICE_LOG("transition to state", "Grouped");
                    return Grouped;
                }
                
                case GroupTrustThisKey: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupTrustThisKey");
                    
                    KeySync_SERVICE_LOG("do action", "trustThisKey");
                    status = trustThisKey(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action trustThisKey() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase2");
                    break;
                }
                
                case GroupKeys: {
                    PEP_STATUS status;
                
                    KeySync_SERVICE_LOG("received event", "GroupKeys");
                    
                    KeySync_SERVICE_LOG("do action", "saveGroupKeys");
                    status = saveGroupKeys(session);
                    if (status == PEP_OUT_OF_MEMORY)
                        return out_of_memory;
                    if (status) {
                        KeySync_ERR_LOG_HEX("executing action saveGroupKeys() failed", status);
                        return invalid_action;
                    }
                
                    KeySync_SERVICE_LOG("remaining in state", "HandshakingGroupedPhase2");
                    break;
                }
                
                default:
                    // ignore events not handled here
                    KeySync_SERVICE_LOG("ignoring event", KeySync_event_name(event));
                    return invalid_event;
            }
            break;
        
        default:
            KeySync_ERR_LOG("invalid state", KeySync_state_name(state));
            return invalid_state;
    }
    
    return None;
}

