// This file is under GNU General Public License 3.0
// see LICENSE.txt

#include "pEp_internal.h"
#include "map_asn1.h"

#include "Sync_impl.h"
#include "KeySync_fsm.h"

static bool _TID_greater(TID_t *t1, TID_t *t2)
{
    assert(t1 && t2);
    if (t1 && !t2)
        return true;
    if (!t1)
        return false;

    if (t1->size > t2->size)
        return true;
    if (t2->size > t1->size)
        return false;

    return memcmp(t1->buf, t2->buf, t1->size) > 0;
}

PEP_STATUS deviceGrouped(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    sqlite3_stmt *_sql;
    int int_result = sqlite3_prepare_v2(session->db,
        "select count(*) from identity where is_own = 1 and (flags & 0x100) = 0x100;"        , -1, &_sql, NULL);
    assert(int_result == SQLITE_OK);
    if (!(int_result == SQLITE_OK))
        return PEP_UNKNOWN_ERROR;

    int _result = 0;
    int_result = Sqlite3_step(_sql);
    assert(int_result == SQLITE_ROW);
    if (int_result == SQLITE_ROW)
        _result = sqlite3_column_int(_sql, 0);
    sqlite3_finalize(_sql);
    if (int_result != SQLITE_ROW)
        return PEP_UNKNOWN_ERROR;

    *result = _result > 0;

    return PEP_STATUS_OK;
}

PEP_STATUS sameChallenge(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    TID_t *t1 = &session->sync_state.keysync.challenge;
    TID_t *t2 = &session->sync_state.own.challenge;

    *result = t1->size == t2->size && memcmp(t1->buf, t2->buf, t1->size) == 0;

    return PEP_STATUS_OK;
}

PEP_STATUS weAreOfferer(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    TID_t *t1 = &session->sync_state.keysync.challenge;
    TID_t *t2 = &session->sync_state.own.challenge;

    *result = _TID_greater(t1, t2);

    return PEP_STATUS_OK;
}

PEP_STATUS sameNegotiation(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    TID_t *t1 = &session->sync_state.keysync.negotiation;
    TID_t *t2 = &session->sync_state.comm_partner.negotiation;

    // test if TID is identical
    *result = t1->size == t2->size && memcmp(t1->buf, t2->buf, t1->size) == 0;

    return PEP_STATUS_OK;
}

PEP_STATUS partnerIsGrouped(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    *result = session->sync_state.keysync.is_group;

    return PEP_STATUS_OK;
}

PEP_STATUS sameNegotiationAndPartner(PEP_SESSION session, bool *result)
{
    assert(session && result);
    if (!(session && result))
        return PEP_ILLEGAL_VALUE;

    TID_t *t1 = &session->sync_state.keysync.negotiation;
    TID_t *t2 = &session->sync_state.comm_partner.negotiation;

    const char *s1 = session->sync_state.comm_partner.sender_fpr;
    const char *s2 = session->sync_state.transport.sender_fpr;

    // test if TID is identical
    *result = t1->size == t2->size && memcmp(t1->buf, t2->buf, t1->size) == 0
    // and test if we're talking to the same sender
            && s1 && s2 && strcmp(s1, s2) == 0;

    return PEP_STATUS_OK;
}

PEP_STATUS newChallengeAndNegotiationBase(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    {
        pEpUUID c;
        uuid_generate_random(c);

        OCTET_STRING_fromBuf(&session->sync_state.own.challenge, (char *) c, 16);
    }
    {
        TID_t *src = &session->sync_state.own.challenge;
        TID_t *dst = &session->sync_state.keysync.challenge;

        assert(src->size == 16);
        if (!(src->size == 16))
            return PEP_UNKNOWN_ERROR;

        OCTET_STRING_fromBuf(dst, (char *) src->buf, src->size);
    }
    {
        pEpUUID c;
        uuid_generate_random(c);

        OCTET_STRING_fromBuf(&session->sync_state.own.negotiation, (char *) c, 16);
    }
    memset(session->sync_state.keysync.negotiation.buf, 0,
            session->sync_state.keysync.negotiation.size);
    memset(session->sync_state.comm_partner.negotiation.buf, 0,
            session->sync_state.comm_partner.negotiation.size);

    return PEP_STATUS_OK;
}

PEP_STATUS showBeingSole(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_SOLE);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS useOwnChallenge(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    {
        TID_t *src = &session->sync_state.own.challenge;
        TID_t *dst = &session->sync_state.keysync.challenge;

        assert(src->size == 16);
        if (!(src->size == 16))
            return PEP_UNKNOWN_ERROR;

        OCTET_STRING_fromBuf(dst, (char *) src->buf, src->size);
    }

    return PEP_STATUS_OK;
}

PEP_STATUS openNegotiation(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    // sender key must be stable while transaction

    // we take the actual signature of the last message and store it in our
    // state for the comm partner
    assert(session->sync_state.transport.sender_fpr);

    free(session->sync_state.comm_partner.sender_fpr);

    session->sync_state.comm_partner.sender_fpr
            = strdup(session->sync_state.transport.sender_fpr);
    assert(session->sync_state.comm_partner.sender_fpr);
    if (!session->sync_state.comm_partner.sender_fpr)
        return PEP_OUT_OF_MEMORY;

    // we need a unique TID for the Negotiation with each single comm_partner
    // we identify the comm_partners by their Challenge
    // we derive the actual Negotiation TID by having random data and XORing it
    // with comm_partner's Challenge

    // copy Negotiation base into buffer

    {
        TID_t *src = &session->sync_state.own.negotiation;
        TID_t *dst = &session->sync_state.keysync.negotiation;

        assert(src->size == 16);
        if (!(src->size == 16))
            return PEP_UNKNOWN_ERROR;

        OCTET_STRING_fromBuf(dst, (char *) src->buf, src->size);
    }

    // we're XORing this with the challenge of the comm_partner, which is in
    // the buffer already

    {
        TID_t *src = &session->sync_state.keysync.challenge;
        TID_t *dst = &session->sync_state.keysync.negotiation;

        assert(src->size == 16 && dst->size == 16);
        if (!(src->size == 16 && dst->size == 16))
            return PEP_UNKNOWN_ERROR;

        for (int i=0; i < src->size; ++i)
            dst->buf[i] ^= src->buf[i];
    }

    // this is the Negotiation's TID for this comm_partner

    {
        TID_t *src = &session->sync_state.keysync.negotiation;
        TID_t *dst = &session->sync_state.comm_partner.negotiation;

        assert(src->size == 16);
        if (!(src->size == 16))
            return PEP_UNKNOWN_ERROR;

        OCTET_STRING_fromBuf(dst, (char *) src->buf, src->size);
    }

    return PEP_STATUS_OK;
}

PEP_STATUS tellWeAreNotGrouped(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    session->sync_state.keysync.is_group = false;

    return PEP_STATUS_OK;
}

PEP_STATUS storeNegotiation(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    // sender key must be stable while transaction

    // we take the actual signature of the last message and store it in our
    // state for the comm partner
    assert(session->sync_state.transport.sender_fpr);

    free(session->sync_state.comm_partner.sender_fpr);

    session->sync_state.comm_partner.sender_fpr
            = strdup(session->sync_state.transport.sender_fpr);
    assert(session->sync_state.comm_partner.sender_fpr);
    if (!session->sync_state.comm_partner.sender_fpr)
        return PEP_OUT_OF_MEMORY;

    {
        TID_t *src = &session->sync_state.keysync.negotiation;
        TID_t *dst = &session->sync_state.comm_partner.negotiation;

        assert(src->size == 16);
        if (!(src->size == 16))
            return PEP_UNKNOWN_ERROR;

        OCTET_STRING_fromBuf(dst, (char *) src->buf, src->size);
    }

    return PEP_STATUS_OK;
}

PEP_STATUS showSoleHandshake(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    assert(session->sync_state.transport.from);
    if (!session->sync_state.transport.from)
        return PEP_ILLEGAL_VALUE;

    pEp_identity *from = session->sync_state.transport.from;
    pEp_identity *me = NULL;
    PEP_STATUS status = get_identity(session, from->address, from->user_id, &me);
    assert(status == PEP_STATUS_OK);
    if (status)
        return status;
    assert(strcmp(me->fpr, session->sync_state.comm_partner.sender_fpr) != 0);
    
    assert(me->fpr && me->fpr[0]);
    if (!(me->fpr && me->fpr[0])) {
        free_identity(me);
        return PEP_ILLEGAL_VALUE;
    }

    pEp_identity *partner = identity_dup(from);
    if (!partner) {
        free_identity(me);
        return PEP_OUT_OF_MEMORY;
    }

    assert(session->sync_state.comm_partner.sender_fpr);
    if (session->sync_state.comm_partner.sender_fpr) {
        free(partner->fpr);
        partner->fpr = strdup(session->sync_state.comm_partner.sender_fpr);
        assert(partner->fpr);
        if (!partner->fpr) {
            free_identity(me);
            free_identity(partner);
            return PEP_OUT_OF_MEMORY;
        }
    }

    free(partner->user_id);
    partner->user_id = strdup("#NV");
    assert(partner->user_id);
    if (!partner->user_id) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    assert(strcmp(me->fpr, partner->fpr) != 0);
    status = session->notifyHandshake(me, partner, SYNC_NOTIFY_INIT_FORM_GROUP);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS disable(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    leave_device_group(session);
    // ignore the result, disable anyway

    if (!session->inject_sync_event)
        return PEP_SYNC_NO_INJECT_CALLBACK;

    session->inject_sync_event(SHUTDOWN, NULL);

    return PEP_STATUS_OK;
}

PEP_STATUS trustThisKey(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->sync_state.transport.from && session->sync_state.comm_partner.sender_fpr);
    if (!(session->sync_state.transport.from && session->sync_state.comm_partner.sender_fpr))
        return PEP_ILLEGAL_VALUE;

    pEp_identity *ident = identity_dup(session->sync_state.transport.from);
    if (!ident)
        return PEP_OUT_OF_MEMORY;
    free(ident->fpr);
    ident->fpr = strdup(session->sync_state.comm_partner.sender_fpr);
    assert(ident->fpr);
    if (!ident->fpr) {
        free_identity(ident);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = trust_own_key(session, ident);
    if (status) {
        free_identity(ident);
        return status;
    }

    OCTET_STRING_fromBuf(&session->sync_state.keysync.key, ident->fpr, strlen(ident->fpr));
    free_identity(ident);

    return PEP_STATUS_OK;
}

PEP_STATUS untrustThisKey(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->sync_state.transport.from && session->sync_state.comm_partner.sender_fpr);
    if (!(session->sync_state.transport.from && session->sync_state.comm_partner.sender_fpr))
        return PEP_ILLEGAL_VALUE;

    pEp_identity *ident = session->sync_state.transport.from;
    free(ident->fpr);
    ident->fpr = strdup(session->sync_state.comm_partner.sender_fpr);
    assert(ident->fpr);
    if (!ident->fpr)
        return PEP_OUT_OF_MEMORY;

    PEP_STATUS status = key_reset_trust(session, ident);
    if (status)
        return status;

    OCTET_STRING_fromBuf(&session->sync_state.keysync.key, "", 0);

    return PEP_STATUS_OK;
}

PEP_STATUS prepareOwnKeys(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    stringlist_t *own_keys;
    PEP_STATUS status = _own_keys_retrieve(session, &own_keys, PEP_idf_not_for_sync, true);
    if (status)
        return status;

    if (session->sync_state.own.keys)
        free_stringlist(session->sync_state.own.keys);
    session->sync_state.own.keys = own_keys;

    identity_list *il;
    status = _own_identities_retrieve(session, &il, PEP_idf_not_for_sync);
    if (status)
        return status;

    IdentityList_from_identity_list(il, &session->sync_state.keysync.ownIdentities);
    free_identity_list(il);

    return PEP_STATUS_OK;
}

PEP_STATUS showFormingGroup(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_FORMING_GROUP);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS saveGroupKeys(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    char *user_id = NULL;
    PEP_STATUS status = get_default_own_userid(session, &user_id);
    if (status)
        return status;

    identity_list *il = IdentityList_to_identity_list(&session->sync_state.keysync.ownIdentities, NULL);
    if (!il) {
        free(user_id);
        return PEP_OUT_OF_MEMORY;
    }
    
    identity_list *oil = session->sync_state.own.identities;

    // BUG: this should be a transaction and been rolled back completely on error
    for (identity_list *_il = il; _il && _il->ident; _il = _il->next) {
        pEp_identity *ident = _il->ident;
        bool is_own_already = false;

        status = is_own_address(session, ident->address, &is_own_already);
        if (status) {
            free_identity_list(il);
            free(user_id);
            return status;
        }

        free(ident->user_id);
        ident->user_id = strdup(user_id);
        assert(ident->user_id);
        if (!ident->user_id) {
            free_identity_list(il);
            free(user_id);
            return PEP_OUT_OF_MEMORY;
        }

        if (is_own_already) {
            ident->comm_type = PEP_ct_pEp;
            status = set_trust(session, ident);
        }
        else {
            oil = identity_list_add(oil, ident);
            if (!oil) {
                free_identity_list(il);
                free(user_id);
                return PEP_OUT_OF_MEMORY;
            }
            status = myself(session, ident);
        }
        if (status) {
            free_identity_list(il);
            free(user_id);
            return status;
        }
    }

    free(user_id);
    free_identity_list(il);

    return PEP_STATUS_OK;
}

PEP_STATUS receivedKeysAreDefaultKeys(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    PEP_STATUS status = PEP_STATUS_OK;
    // set flag for all keys
    for (identity_list *il = session->sync_state.own.identities; il && il->ident ; il = il->next) {
        if (!(il->ident->flags && PEP_idf_not_for_sync)) {
            
            status = set_identity_flags(session, il->ident, PEP_idf_devicegroup);
            if (status)
                return status;
        }
    }

    char *user_id = NULL;
    status = get_default_own_userid(session, &user_id);
    if (status)
        return status;

    identity_list *il = IdentityList_to_identity_list(&session->sync_state.keysync.ownIdentities, NULL);
    if (!il)
        return PEP_OUT_OF_MEMORY;

    for (identity_list *_il = il; _il && _il->ident ; _il = _il->next) {
        // replace partner's user_id with own user_id
        free(_il->ident->user_id);
        _il->ident->user_id = strdup(user_id);
        assert(_il->ident->user_id);
        if (!_il->ident->user_id) {
            free_identity_list(il);
            free(user_id);
            return PEP_OUT_OF_MEMORY;
        }
    
        const char *own_key = _il->ident->fpr;
        _il->ident->fpr = NULL;
        status = set_own_key(session, _il->ident, own_key);
        if (status) {
            free_identity_list(il);
            free(user_id);
            return status;
        }
    }

    free_identity_list(il);
    free(user_id);

    return PEP_STATUS_OK;
}

PEP_STATUS showGroupCreated(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_ACCEPTED_GROUP_CREATED);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS ownKeysAreDefaultKeys(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    PEP_STATUS status = PEP_STATUS_OK;

    // set flag for all keys; don't change anything else
    for (identity_list *il = session->sync_state.own.identities; il && il->ident ; il = il->next) {
        if (!(il->ident->flags && PEP_idf_not_for_sync)) {
            status = set_identity_flags(session, il->ident, PEP_idf_devicegroup);
            if (status)
                return status;
        }
    }

    return PEP_STATUS_OK;
}

PEP_STATUS showBeingInGroup(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_IN_GROUP);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS tellWeAreGrouped(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    session->sync_state.keysync.is_group = true;

    return PEP_STATUS_OK;
}

PEP_STATUS showJoinGroupHandshake(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    assert(session->sync_state.transport.from);
    if (!session->sync_state.transport.from)
        return PEP_ILLEGAL_VALUE;

    pEp_identity *from = session->sync_state.transport.from;
    pEp_identity *me = NULL;
    PEP_STATUS status = get_identity(session, from->address, from->user_id, &me);
    assert(status == PEP_STATUS_OK);
    if (status)
        return status;
    assert(strcmp(me->fpr, session->sync_state.comm_partner.sender_fpr) != 0);
    
    assert(me->fpr && me->fpr[0]);
    if (!(me->fpr && me->fpr[0])) {
        free_identity(me);
        return PEP_ILLEGAL_VALUE;
    }

    pEp_identity *partner = identity_dup(from);
    if (!partner) {
        free_identity(me);
        return PEP_OUT_OF_MEMORY;
    }

    assert(session->sync_state.comm_partner.sender_fpr);
    if (session->sync_state.comm_partner.sender_fpr) {
        free(partner->fpr);
        partner->fpr = strdup(session->sync_state.comm_partner.sender_fpr);
        assert(partner->fpr);
        if (!partner->fpr) {
            free_identity(me);
            free_identity(partner);
            return PEP_OUT_OF_MEMORY;
        }
    }

    free(partner->user_id);
    partner->user_id = strdup("#NV");
    assert(partner->user_id);
    if (!partner->user_id) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    assert(strcmp(me->fpr, partner->fpr) != 0);
    status = session->notifyHandshake(me, partner, SYNC_NOTIFY_INIT_ADD_OUR_DEVICE);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS showDeviceAdded(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_ACCEPTED_DEVICE_ADDED);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS showGroupedHandshake(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    assert(session->sync_state.transport.from);
    if (!session->sync_state.transport.from)
        return PEP_ILLEGAL_VALUE;

    pEp_identity *from = session->sync_state.transport.from;
    pEp_identity *me = NULL;
    PEP_STATUS status = get_identity(session, from->address, from->user_id, &me);
    assert(status == PEP_STATUS_OK);
    if (status)
        return status;
    assert(strcmp(me->fpr, session->sync_state.comm_partner.sender_fpr) != 0);
    
    assert(me->fpr && me->fpr[0]);
    if (!(me->fpr && me->fpr[0])) {
        free_identity(me);
        return PEP_ILLEGAL_VALUE;
    }

    pEp_identity *partner = identity_dup(from);
    if (!partner) {
        free_identity(me);
        return PEP_OUT_OF_MEMORY;
    }

    assert(session->sync_state.comm_partner.sender_fpr);
    if (session->sync_state.comm_partner.sender_fpr) {
        free(partner->fpr);
        partner->fpr = strdup(session->sync_state.comm_partner.sender_fpr);
        assert(partner->fpr);
        if (!partner->fpr) {
            free_identity(me);
            free_identity(partner);
            return PEP_OUT_OF_MEMORY;
        }
    }

    free(partner->user_id);
    partner->user_id = strdup("#NV");
    assert(partner->user_id);
    if (!partner->user_id) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    assert(strcmp(me->fpr, partner->fpr) != 0);
    status = session->notifyHandshake(me, partner, SYNC_NOTIFY_INIT_ADD_OTHER_DEVICE);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS hideHandshakeDialog(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_OVERTAKEN);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

PEP_STATUS KeySyncTimeoutHandler(PEP_SESSION session)
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    assert(session->notifyHandshake);
    if (!session->notifyHandshake)
        return PEP_SYNC_NO_NOTIFY_CALLBACK;
 
    pEp_identity *me = new_identity(NULL, NULL, NULL, NULL);
    pEp_identity *partner = new_identity(NULL, NULL, NULL, NULL);
    assert(me && partner);
    if (!(me && partner)) {
        free_identity(me);
        free_identity(partner);
        return PEP_OUT_OF_MEMORY;
    }

    PEP_STATUS status = session->notifyHandshake(me, partner, SYNC_NOTIFY_TIMEOUT);
    if (status)
        return status;

    return PEP_STATUS_OK;
}

