// This file is under GNU General Public License 3.0
// see LICENSE.txt

#include <assert.h>
#include <stdlib.h>
#include "pEp_internal.h"
#include "map_asn1.h"
#include "Sync_func.h"
#include "KeySync_fsm.h"

void free_Sync_state(PEP_SESSION session)
{
    if (!session)
        return;

    // own state

    free_stringlist(session->sync_state.own.keys);
    session->sync_state.own.keys = NULL;
    free_identity_list(session->sync_state.own.identities);
    session->sync_state.own.identities = NULL;

    // TIDs we're using ourselves
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.own.challenge);
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.own.negotiation);

    // state we learned about our communication partner

    free(session->sync_state.comm_partner.sender_fpr);
    session->sync_state.comm_partner.sender_fpr = NULL;

    // TIDs our comm partner wants to have
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.comm_partner.challenge);
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.comm_partner.negotiation);

    // buffer for transport data

    free_identity(session->sync_state.transport.from);
    session->sync_state.transport.from = NULL;
    free(session->sync_state.transport.sender_fpr);
    session->sync_state.transport.sender_fpr = NULL;

    // message buffers

    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.keysync.challenge);
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_TID, &session->sync_state.keysync.negotiation);
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_Hash, &session->sync_state.keysync.key);
    ASN_STRUCT_FREE_CONTENTS_ONLY(asn_DEF_IdentityList, &session->sync_state.keysync.ownIdentities);

    memset(&session->sync_state, 0, sizeof(session->sync_state));
}

Sync_t *new_Sync_message(Sync_PR fsm, int message_type)
{
    Sync_t *msg = calloc(sizeof(Sync_t), 1);
    assert(msg);
    if (!msg)
        return NULL;

    if (fsm) {
        msg->present = fsm;
        if (message_type) {
            switch (fsm) {
                case Sync_PR_keysync:
                    msg->choice.keysync.present = message_type;
                        break;
                
                default:
                    free(msg);
                    return NULL;
            }
        }
    }

    return msg;
}

void free_Sync_message(Sync_t *msg)
{
    ASN_STRUCT_FREE(asn_DEF_Sync, msg);
}

PEP_STATUS update_Sync_state(PEP_SESSION session, Sync_t *msg,
        Sync_PR *fsm, int *message_type)
{
    int result = 0;

    assert(session && msg && fsm && message_type);
    if (!(session && msg && fsm && message_type))
        return PEP_ILLEGAL_VALUE;

    *fsm = 0;
    *message_type = None;

    switch (msg->present) {
        case Sync_PR_NOTHING:
            return PEP_ILLEGAL_VALUE;

        case Sync_PR_keysync:
            switch (msg->choice.keysync.present) {
                case KeySync_PR_NOTHING:
                    return PEP_ILLEGAL_VALUE;
        
                case KeySync_PR_beacon:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.challenge,
                            (char *) msg->choice.keysync.choice.beacon.challenge.buf,
                            msg->choice.keysync.choice.beacon.challenge.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = Beacon;
                    break;
                
                case KeySync_PR_negotiationRequest:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.challenge,
                            (char *) msg->choice.keysync.choice.negotiationRequest.challenge.buf,
                            msg->choice.keysync.choice.negotiationRequest.challenge.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.negotiationRequest.negotiation.buf,
                            msg->choice.keysync.choice.negotiationRequest.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    session->sync_state.keysync.is_group = msg->choice.keysync
                            .choice.negotiationRequest.is_group;
                    
                    *message_type = NegotiationRequest;
                    break;
                
                case KeySync_PR_negotiationOpen:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.negotiationOpen.negotiation.buf,
                            msg->choice.keysync.choice.negotiationOpen.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = NegotiationOpen;
                    break;
                
                case KeySync_PR_rollback:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.rollback.negotiation.buf,
                            msg->choice.keysync.choice.rollback.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = Rollback;
                    break;
                
                case KeySync_PR_commitReject:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.commitReject.negotiation.buf,
                            msg->choice.keysync.choice.commitReject.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = CommitReject;
                    break;
                
                case KeySync_PR_commitAcceptOfferer:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.commitAcceptOfferer.negotiation.buf,
                            msg->choice.keysync.choice.commitAcceptOfferer.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = CommitAcceptOfferer;
                    break;
                
                case KeySync_PR_commitAcceptRequester:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.commitAcceptRequester.negotiation.buf,
                            msg->choice.keysync.choice.commitAcceptRequester.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = CommitAcceptRequester;
                    break;
                
                case KeySync_PR_commitAccept:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.commitAccept.negotiation.buf,
                            msg->choice.keysync.choice.commitAccept.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = CommitAccept;
                    break;
                
                case KeySync_PR_commitAcceptForGroup:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.negotiation,
                            (char *) msg->choice.keysync.choice.commitAcceptForGroup.negotiation.buf,
                            msg->choice.keysync.choice.commitAcceptForGroup.negotiation.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = CommitAcceptForGroup;
                    break;
                
                case KeySync_PR_groupTrustThisKey:
                    result = OCTET_STRING_fromBuf(&session->sync_state.keysync.key,
                            (char *) msg->choice.keysync.choice.groupTrustThisKey.key.buf,
                            msg->choice.keysync.choice.groupTrustThisKey.key.size);
                    if (result)
                        return PEP_OUT_OF_MEMORY;
                    
                    *message_type = GroupTrustThisKey;
                    break;
                
                case KeySync_PR_groupKeys:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &msg->choice.keysync.choice.groupKeys.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &session->sync_state.keysync.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    *message_type = GroupKeys;
                    break;
                
                case KeySync_PR_ownKeysOfferer:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &msg->choice.keysync.choice.ownKeysOfferer.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &session->sync_state.keysync.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    *message_type = OwnKeysOfferer;
                    break;
                
                case KeySync_PR_ownKeysRequester:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &msg->choice.keysync.choice.ownKeysRequester.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &session->sync_state.keysync.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    *message_type = OwnKeysRequester;
                    break;
                
                default:
                    return PEP_ILLEGAL_VALUE;
            }
            break;
        
        default:
            return PEP_ILLEGAL_VALUE;
    }

    *fsm = msg->present;
    return PEP_STATUS_OK;
}

PEP_STATUS update_Sync_message(PEP_SESSION session, Sync_t *msg)
{
    assert(session && msg);
    if (!(session && msg))
        return PEP_ILLEGAL_VALUE;

    int fsm = msg->present;
    switch (fsm) {
        case Sync_PR_NOTHING:
            return PEP_ILLEGAL_VALUE;

        case Sync_PR_keysync:
            {
                int message_type = msg->choice.keysync.present;
                switch (message_type) {
                    case KeySync_PR_NOTHING:
                        return PEP_ILLEGAL_VALUE;
        
                case KeySync_PR_beacon:
                    {
                        long *major = (long *) malloc(sizeof(long));
                        long *minor = (long *) malloc(sizeof(long));
                        assert(major && minor);
                        if (!(major && minor))
                            return PEP_OUT_OF_MEMORY;
                    
                        *major = 1;
                        *minor = 2;
                    
                        msg->choice.keysync.choice.beacon.version.major = major;
                        msg->choice.keysync.choice.beacon.version.minor = minor;
                    }
                    
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.beacon.challenge,
                                (char *) session->sync_state.keysync.challenge.buf,
                                session->sync_state.keysync.challenge.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_negotiationRequest:
                    {
                        long *major = (long *) malloc(sizeof(long));
                        long *minor = (long *) malloc(sizeof(long));
                        assert(major && minor);
                        if (!(major && minor))
                            return PEP_OUT_OF_MEMORY;
                    
                        *major = 1;
                        *minor = 2;
                    
                        msg->choice.keysync.choice.negotiationRequest.version.major = major;
                        msg->choice.keysync.choice.negotiationRequest.version.minor = minor;
                    }
                    
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.negotiationRequest.challenge,
                                (char *) session->sync_state.keysync.challenge.buf,
                                session->sync_state.keysync.challenge.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.negotiationRequest.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    msg->choice.keysync.choice.negotiationRequest.is_group
                             = session->sync_state.keysync.is_group;
                    
                    break;
                
                case KeySync_PR_negotiationOpen:
                    {
                        long *major = (long *) malloc(sizeof(long));
                        long *minor = (long *) malloc(sizeof(long));
                        assert(major && minor);
                        if (!(major && minor))
                            return PEP_OUT_OF_MEMORY;
                    
                        *major = 1;
                        *minor = 2;
                    
                        msg->choice.keysync.choice.negotiationOpen.version.major = major;
                        msg->choice.keysync.choice.negotiationOpen.version.minor = minor;
                    }
                    
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.negotiationOpen.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_rollback:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.rollback.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_commitReject:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.commitReject.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_commitAcceptOfferer:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.commitAcceptOfferer.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_commitAcceptRequester:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.commitAcceptRequester.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_commitAccept:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.commitAccept.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_commitAcceptForGroup:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.commitAcceptForGroup.negotiation,
                                (char *) session->sync_state.keysync.negotiation.buf,
                                session->sync_state.keysync.negotiation.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_groupTrustThisKey:
                    {
                        int result = OCTET_STRING_fromBuf(
                                &msg->choice.keysync.choice.groupTrustThisKey.key,
                                (char *) session->sync_state.keysync.key.buf,
                                session->sync_state.keysync.key.size
                            );
                        if (result)
                            return PEP_OUT_OF_MEMORY;
                    }
                    break;
                
                case KeySync_PR_groupKeys:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &session->sync_state.keysync.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &msg->choice.keysync.choice.groupKeys.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    break;
                
                case KeySync_PR_ownKeysOfferer:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &session->sync_state.keysync.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &msg->choice.keysync.choice.ownKeysOfferer.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    break;
                
                case KeySync_PR_ownKeysRequester:
                    {
                        identity_list *il = IdentityList_to_identity_list(
                                &session->sync_state.keysync.ownIdentities, NULL);
                        if (!il)
                            return PEP_OUT_OF_MEMORY;
                        IdentityList_t *_il = IdentityList_from_identity_list(il,
                                &msg->choice.keysync.choice.ownKeysRequester.ownIdentities);
                        free_identity_list(il);
                        if (!_il)
                            return PEP_OUT_OF_MEMORY;
                    }
                    
                    break;
                
                    default:
                        return PEP_ILLEGAL_VALUE;
                }
                break;
            }
        default:
            return PEP_ILLEGAL_VALUE;
    }

    return PEP_STATUS_OK;
}

