// This file is under GNU General Public License 3.0
// see LICENSE.txt

#include "Sync_impl.h"
#include "pEp_internal.h"
#include "Sync_event.h"
#include "sync_codec.h"
#include "baseprotocol.h"
#include "KeySync_fsm.h"

static bool _KeySync_timeout(int state)
{
    static int last_state = None;
    static time_t switch_time = 0;

    if (state > Init) {
        if (state == last_state) {
            if (time(NULL) - switch_time > KEYSYNC_THRESHOLD) {
                last_state = None;
                switch_time = 0;
                return true;
            }
        }
        else {
            last_state = state;
            switch_time = time(NULL);
        }
    }
    else {
        last_state = None;
        switch_time = 0;
    }

    return false;
}

PEP_STATUS Sync_driver(
        PEP_SESSION session,
        Sync_PR fsm,
        int event
    )
{
    assert(session);
    if (!session)
        return PEP_ILLEGAL_VALUE;

    switch (fsm) {
        case None:
            if (!event) {
                // timeout occured
                Sync_driver(session, Sync_PR_keysync, None);
                return PEP_STATUS_OK;
            }
            return PEP_ILLEGAL_VALUE;

        case Sync_PR_keysync: {
            int state = session->sync_state.keysync.state;
            switch (state) {
                case HandshakingOfferer:
                case HandshakingRequester:
                case HandshakingPhase1Offerer:
                case HandshakingPhase1Requester:
                case HandshakingPhase2Offerer:
                case HandshakingPhase2Requester:
                case FormingGroupOfferer:
                case FormingGroupRequester:
                case HandshakingToJoin:
                case HandshakingToJoinPhase1:
                case HandshakingToJoinPhase2:
                case JoiningGroup:
                case HandshakingGrouped:
                case HandshakingGroupedPhase1:
                case HandshakingGroupedPhase2:
                    if (_KeySync_timeout(state)) {
                        session->sync_state.keysync.state = Init;
                        event = Init;
                        KeySyncTimeoutHandler(session);
                    }
                    break;
                
                default:
                    _KeySync_timeout(None);
            }
            break;
        }
    
        default:
            return PEP_ILLEGAL_VALUE;
    }

    int next_state = None;
    do {
        switch (fsm) {
            case Sync_PR_keysync: {
                int state = session->sync_state.keysync.state;
                next_state = fsm_KeySync(session, state, event);
                if (next_state > None) {
                    session->sync_state.keysync.state = next_state;
                    event = Init;
                }
                else if (next_state < None) {
                    return PEP_STATEMACHINE_ERROR - next_state;
                }
                break;
            }
            
            default:
                return PEP_ILLEGAL_VALUE;
        }
    }  while (next_state);

    return PEP_STATUS_OK;
}

PEP_STATUS signal_Sync_event(
        PEP_SESSION session, 
        Sync_PR fsm,
        int event,
        identity_list *own_identities
    )
{
    Sync_t *msg = NULL;
    Sync_event_t *ev = NULL;

    assert(session && fsm > 0 && event > None);
    if (!(session && fsm > 0 && event > None))
        return PEP_ILLEGAL_VALUE;

    PEP_STATUS status = PEP_STATUS_OK;

    if (!session->inject_sync_event)
       return PEP_SYNC_NO_INJECT_CALLBACK;

    if (event < Extra) {
        msg = new_Sync_message(fsm, event);
        if (!msg) {
            status = PEP_OUT_OF_MEMORY;
            goto the_end;
        }

        status = update_Sync_message(session, msg);
        if (status)
            goto the_end;
    }

    ev = new_Sync_event(fsm, event, msg);
    if (!ev) {
        status = PEP_OUT_OF_MEMORY;
        goto the_end;
    }

    ev->own_identities = own_identities;

    int result = session->inject_sync_event(ev,
            session->sync_management);
    if (result) {
        status = PEP_STATEMACHINE_ERROR;
        goto the_end;
    }
    return PEP_STATUS_OK;

the_end:
    free_Sync_event(ev); // msg gets freed here
    return status;
}

PEP_STATUS signal_Sync_message(
        PEP_SESSION session, 
        PEP_rating rating,
        const char *data,
        size_t size,
        const pEp_identity *from,
        const char *sender_fpr
    )
{
    assert(session && data && size);
    if (!(session && data && size))
        return PEP_ILLEGAL_VALUE;

    if (!session->inject_sync_event)
       return PEP_SYNC_NO_INJECT_CALLBACK;

    PEP_STATUS status = PEP_STATUS_OK;
    Sync_event_t *ev = NULL;

    Sync_t *msg = NULL;
    status = decode_Sync_message(data, size, &msg);
    if (status)
        return status;

    Sync_PR fsm = msg->present;
    int event = 0;
    bool is_own_key = false;

    switch (fsm) {
        case Sync_PR_keysync:
            switch (msg->choice.keysync.present) {
                // these messages require a detached signature
                case KeySync_PR_beacon:
                    if (!sender_fpr) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = Beacon;
                    break;
        
                // these messages must arrive encrypted
                case KeySync_PR_negotiationRequest:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = NegotiationRequest;
                    break;
        
                case KeySync_PR_negotiationOpen:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = NegotiationOpen;
                    break;
        
                case KeySync_PR_rollback:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = Rollback;
                    break;
        
                case KeySync_PR_commitReject:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = CommitReject;
                    break;
        
                case KeySync_PR_commitAcceptOfferer:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = CommitAcceptOfferer;
                    break;
        
                case KeySync_PR_commitAcceptRequester:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = CommitAcceptRequester;
                    break;
        
                case KeySync_PR_commitAccept:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = CommitAccept;
                    break;
        
                case KeySync_PR_commitAcceptForGroup:
                    if (rating < PEP_rating_reliable) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = CommitAcceptForGroup;
                    break;
        
                // these messages must come through a trusted channel
                case KeySync_PR_groupTrustThisKey:
                    if (rating < PEP_rating_trusted) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    status = own_key_is_listed(session, sender_fpr, &is_own_key);
                    if (status)
                        goto the_end;
                    if (!is_own_key) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = GroupTrustThisKey;
                    break;
        
                case KeySync_PR_groupKeys:
                    if (rating < PEP_rating_trusted) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    status = own_key_is_listed(session, sender_fpr, &is_own_key);
                    if (status)
                        goto the_end;
                    if (!is_own_key) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = GroupKeys;
                    break;
        
                case KeySync_PR_ownKeysOfferer:
                    if (rating < PEP_rating_trusted) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    status = own_key_is_listed(session, sender_fpr, &is_own_key);
                    if (status)
                        goto the_end;
                    if (!is_own_key) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = OwnKeysOfferer;
                    break;
        
                case KeySync_PR_ownKeysRequester:
                    if (rating < PEP_rating_trusted) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    status = own_key_is_listed(session, sender_fpr, &is_own_key);
                    if (status)
                        goto the_end;
                    if (!is_own_key) {
                        status = PEP_SYNC_ILLEGAL_MESSAGE;
                        goto the_end;
                    }
                    event = OwnKeysRequester;
                    break;
        
                default:
                    status = PEP_SYNC_ILLEGAL_MESSAGE;
                    goto the_end;
            }
            break;
        
        default:
            status = PEP_SYNC_ILLEGAL_MESSAGE;
            goto the_end;
    }

    ev = new_Sync_event(fsm, event, msg);
    if (!ev) {
        status = PEP_OUT_OF_MEMORY;
        goto the_end;
    }

    // add transport data

    if (from) {
        ev->from = identity_dup(from);
        if (!ev->from) {
            status = PEP_OUT_OF_MEMORY;
            goto the_end;
        }
    }

    if (sender_fpr) {
        ev->sender_fpr = strdup(sender_fpr);
        assert(ev->sender_fpr);
        if (!ev->sender_fpr) {
            status = PEP_OUT_OF_MEMORY;
            goto the_end;
        }
    }

    int result = session->inject_sync_event(ev,
            session->sync_management);
    if (result) {
        status = PEP_STATEMACHINE_ERROR;
        goto the_end;
    }

    return PEP_STATUS_OK;

the_end:
    free_Sync_event(ev); // msg gets freed here
    return status;
}

PEP_STATUS send_Sync_message(
        PEP_SESSION session, 
        Sync_PR fsm,
        int message_type
    )
{
    PEP_STATUS status = PEP_STATUS_OK;

    assert(session && fsm > None && message_type > None);
    if (!(session && fsm > None && message_type > None))
        return PEP_ILLEGAL_VALUE;
    
    Sync_t *msg = new_Sync_message(fsm, message_type);
    if (!msg)
        return PEP_OUT_OF_MEMORY;

    char *data = NULL;
    message *m = NULL;
    identity_list *channels = NULL;
    char *key_data = NULL;
    size_t key_data_size = 0;
    stringlist_t *extra = NULL;
    bool transaction;

    status = update_Sync_message(session, msg);
    if (status)
        goto the_end;

    size_t size = 0;
    status = encode_Sync_message(msg, &data, &size);
    if (status)
        goto the_end;

    switch (message_type) {
        // these messages are being broadcasted
        case KeySync_PR_beacon:
            status = _own_identities_retrieve(session, &channels, PEP_idf_not_for_sync);
            if (status)
                goto the_end;

            if (!(channels && channels->ident)) {
                // status = PEP_SYNC_NO_CHANNEL;
                // we don't check for having a channel, because if
                // this is initial setup before having an own
                // identity we're fine
                goto the_end;
            }
            break;

        // these go anycast; previously used address is sticky (unicast)
        case KeySync_PR_negotiationRequest:
        case KeySync_PR_negotiationOpen:
        case KeySync_PR_rollback:
        case KeySync_PR_commitReject:
        case KeySync_PR_commitAcceptOfferer:
        case KeySync_PR_commitAcceptRequester:
        case KeySync_PR_commitAccept:
        case KeySync_PR_commitAcceptForGroup:
        case KeySync_PR_groupTrustThisKey:
        case KeySync_PR_groupKeys:
        case KeySync_PR_ownKeysOfferer:
        case KeySync_PR_ownKeysRequester:
            if (!session->sync_state.transport.from ||
                (session->sync_state.transport.from->flags &
                PEP_idf_not_for_sync)) {

                // no address available yet, try to find one
                status = _own_identities_retrieve(session, &channels, PEP_idf_not_for_sync);
                if (status)
                    goto the_end;
                break;

                if (channels && channels->ident) {
                    // only need the first one
                    free_identity_list(channels->next);
                    channels->next = NULL;
                }
                else {
                    status = PEP_SYNC_NO_CHANNEL;
                    goto the_end;
                }
            }
            else {
                pEp_identity *channel = identity_dup(session->sync_state.transport.from);
                if (!channel) {
                    status = PEP_OUT_OF_MEMORY;
                    goto the_end;
                }

                channels = new_identity_list(channel);
                if (!channels) {
                    status = PEP_OUT_OF_MEMORY;
                    goto the_end;
                }
            }
            break;

        default:
            status = PEP_SYNC_ILLEGAL_MESSAGE;
            goto the_end;
    }

    for (identity_list *li = channels; li && li->ident ; li = li->next) {
        message *_m = NULL;
        char *_data = NULL;
        
        _data = malloc(size);
        assert(_data);
        if (!_data) {
            status = PEP_OUT_OF_MEMORY;
            goto the_end;
        }
        memcpy(_data, data, size);

        switch (message_type) {
            case KeySync_PR_beacon:
                status = base_prepare_message(
                        session,
                        li->ident,
                        li->ident,
                        _data,
                        size,
                        li->ident->fpr,
                        &_m
                    );
                if (status) {
                    free(_data);
                    goto the_end;
                }
                attach_own_key(session, _m);
                decorate_message(_m, PEP_rating_undefined, NULL, true, true);
                m = _m;
                break;

            case KeySync_PR_negotiationRequest:
            case KeySync_PR_negotiationOpen:
            case KeySync_PR_rollback:
            case KeySync_PR_commitReject:
            case KeySync_PR_commitAcceptOfferer:
            case KeySync_PR_commitAcceptRequester:
            case KeySync_PR_commitAccept:
            case KeySync_PR_commitAcceptForGroup:
                // add fpr of key of comm partner

                assert(session->sync_state.comm_partner.sender_fpr);
                if (!session->sync_state.comm_partner.sender_fpr) {
                    status = PEP_SYNC_CANNOT_ENCRYPT;
                    goto the_end;
                }

                extra = new_stringlist(session->sync_state.comm_partner.sender_fpr);
                if (!extra) {
                    status = PEP_OUT_OF_MEMORY;
                    goto the_end;
                }

                status = base_prepare_message(
                        session,
                        li->ident,
                        li->ident,
                        _data,
                        size,
                        NULL,
                        &_m
                    );
                if (status) {
                    free(_data);
                    goto the_end;
                }

                status = encrypt_message(session, _m, extra, &m, PEP_enc_PEP, 0);
                if (status) {
                    status = PEP_SYNC_CANNOT_ENCRYPT;
                    goto the_end;
                }
                add_opt_field(m, "pEp-auto-consume", "yes");
                add_opt_field(m, "in-reply-to:", "pEp-auto-consume@pEp.foundation");
                free_message(_m);
                break;

            case KeySync_PR_groupKeys:
            case KeySync_PR_ownKeysOfferer:
            case KeySync_PR_ownKeysRequester:
                // check if we had a former negotiation

                transaction = false;
                for (int i=0; i < session->sync_state.own.negotiation.size; i++) {
                    if (session->sync_state.own.negotiation.buf[i]) {
                        transaction = true;
                        break;
                    }
                }

                // if it is a former negotiation check if the key
                // is fully trusted and the sender key of this
                // transaction; if so add the sender key to extra
                // keys allowing this new partner to read the
                // secret keys

                if (transaction) {
                    assert(session->sync_state.comm_partner.sender_fpr &&
                        session->sync_state.transport.from &&
                        session->sync_state.transport.from->user_id);
                    if (!(session->sync_state.comm_partner.sender_fpr &&
                            session->sync_state.transport.from &&
                            session->sync_state.transport.from->user_id))
                    {
                        status = PEP_SYNC_CANNOT_ENCRYPT;
                        goto the_end;
                    }
                
                    // test if this is a green channel

                    pEp_identity *ident = new_identity(NULL,
                            session->sync_state.comm_partner.sender_fpr,
                            session->sync_state.transport.from->user_id,
                            NULL
                        );
                    if (!ident) {
                        status = PEP_OUT_OF_MEMORY;
                        goto the_end;
                    }
                    status = get_trust(session, ident);
                    if (status) {
                        free_identity(ident);
                        goto the_end;
                    }
                    assert(ident->comm_type == PEP_ct_pEp); // we don't deliver otherwise
                    if (ident->comm_type != PEP_ct_pEp) {
                        free_identity(ident);
                        status = PEP_SYNC_CANNOT_ENCRYPT;
                        goto the_end;
                    }
                    free_identity(ident);

                    // test if we accepted this as own key already

                    bool is_own_key = false;
                    status = own_key_is_listed(session,
                            session->sync_state.comm_partner.sender_fpr,
                            &is_own_key);
                    assert(!status);
                    if (status)
                        goto the_end;
                    assert(is_own_key);
                    if (!is_own_key) {
                        status = PEP_SYNC_CANNOT_ENCRYPT;
                        goto the_end;
                    }

                    // if so add key of comm partner to extra keys

                    extra = new_stringlist(session->sync_state.comm_partner.sender_fpr);
                    if (!extra) {
                        status = PEP_OUT_OF_MEMORY;
                        goto the_end;
                    }
                }
                
                status = base_prepare_message(
                        session,
                        li->ident,
                        li->ident,
                        _data,
                        size,
                        NULL,
                        &_m
                    );
                if (status) {
                    free(_data);
                    goto the_end;
                }

                // export secret keys into memory

                key_data = strdup("");
                assert(key_data);
                if (!key_data) {
                    free(_data);
                    free_message(_m);
                    status = PEP_OUT_OF_MEMORY;
                    goto the_end;
                }
                key_data_size = 1; // N.B. If null termination makes us happy for debugging, fine, but 
                                   // if we include this in the size, libetpan will null terminate and 
                                   // go bananas. We can't have a NUL in the mime text.

                for (stringlist_t *sl = session->sync_state.own.keys;
                        sl && sl->value ; sl = sl->next)
                {
                    char *_key_data = NULL;
                    size_t _size = 0;
                    status = export_secret_key(session, sl->value, &_key_data, &_size);
                    if (status && status != PEP_KEY_NOT_FOUND) {
                        free(_data);
                        free_message(_m);
                        goto the_end;
                    }

                    if (status != PEP_KEY_NOT_FOUND) {
                        assert(_key_data && _size);
                        char *n = realloc(key_data, key_data_size + _size);
                        if (!n) {
                            free(_data);
                            free_message(_m);
                            status = PEP_OUT_OF_MEMORY;
                            goto the_end;
                        }
                        key_data = n;
                        key_data_size += _size;
                        strlcat(key_data, _key_data, key_data_size);
                        free(_key_data);
                        _key_data = NULL;
                    }
                    status = export_key(session, sl->value, &_key_data, &_size);
                    if (status && status != PEP_KEY_NOT_FOUND) {
                        free(_data);
                        free_message(_m);
                        goto the_end;
                    }

                    if (status != PEP_KEY_NOT_FOUND) {
                        assert(_key_data && _size);
                        char *n = realloc(key_data, key_data_size + _size);
                        if (!n) {
                            free(_data);
                            free_message(_m);
                            status = PEP_OUT_OF_MEMORY;
                            goto the_end;
                        }
                        key_data = n;
                        key_data_size += _size;
                        strlcat(key_data, _key_data, key_data_size);
                        free(_key_data);
                        _key_data = NULL;
                    }
                }

                // add secret key data as attachment

                // N.B. The -1 makes sure we do NOT add a NUL into the mime stream!
                bloblist_t *bl = bloblist_add(_m->attachments, key_data, key_data_size - 1,
                        "application/octet-stream", "file://own.key");
                if (!bl) {
                    free(_data);
                    free_message(_m);
                    status = PEP_OUT_OF_MEMORY;
                    goto the_end;
                }
                key_data = NULL;

                status = encrypt_message(session, _m, extra, &m, PEP_enc_PEP, 0);
                if (status) {
                    status = PEP_SYNC_CANNOT_ENCRYPT;
                    goto the_end;
                }
                add_opt_field(m, "pEp-auto-consume", "yes");
                add_opt_field(m, "in-reply-to:", "pEp-auto-consume@pEp.foundation");
                free_message(_m);
                break;

            default: // security=trusted only
                status = base_prepare_message(
                        session,
                        li->ident,
                        li->ident,
                        _data,
                        size,
                        NULL,
                        &_m
                    );
                if (status) {
                    free(_data);
                    goto the_end;
                }

                status = encrypt_message(session, _m, NULL, &m, PEP_enc_PEP, 0);
                if (status) {
                    status = PEP_SYNC_CANNOT_ENCRYPT;
                    goto the_end;
                }
                add_opt_field(m, "pEp-auto-consume", "yes");
                add_opt_field(m, "in-reply-to:", "pEp-auto-consume@pEp.foundation");
                free_message(_m);
        }

        status = session->messageToSend(m);
        m = NULL;
    }

the_end:
    free_stringlist(extra);
    free_identity_list(channels);
    free_message(m);
    free(data);
    free(key_data);
    free_Sync_message(msg);
    if (status)
        SERVICE_ERROR_LOG(session, "send_Sync_message()", status);
    return status;
}

PEP_STATUS recv_Sync_event(
        PEP_SESSION session,
        Sync_event_t *ev
    )
{
    assert(session && ev);
    if (!(session && ev))
        return PEP_ILLEGAL_VALUE;

    PEP_STATUS status = PEP_STATUS_OK;
    Sync_PR fsm = (int) None;
    int event = None;

    if (ev->event > None && ev->event < Extra) {
        status = update_Sync_state(session, ev->msg, &fsm, &event);
        if (status)
            goto the_end;

        if (ev->fsm) {
            if (ev->fsm != fsm || ev->event != event) {
                status = PEP_SYNC_ILLEGAL_MESSAGE;
                goto the_end;
            }
        }
        else if (ev->event) {
            status = PEP_SYNC_ILLEGAL_MESSAGE;
            goto the_end;
        }
    }
    else {
        fsm = ev->fsm;
        event = ev->event;
    }

    // update transport data

    if (ev->from) {
        free_identity(session->sync_state.transport.from);
        session->sync_state.transport.from = ev->from;
        ev->from = NULL;
    }

    if (ev->sender_fpr) {
        free(session->sync_state.transport.sender_fpr);
        session->sync_state.transport.sender_fpr = ev->sender_fpr;
        ev->sender_fpr = NULL;
    }

    // update own identities

    if (ev->own_identities && ev->own_identities->ident) {
        free_identity_list(session->sync_state.own.identities);
        session->sync_state.own.identities = ev->own_identities;
        ev->own_identities = NULL;
    }

    status = Sync_driver(session, fsm, event);

the_end:
    //free_Sync_event(ev); // FIXME: We don't own this pointer. Are we sure it gets freed externally?
    return status;
}

